@@ -1,29 +1,8 @@
-/*
-============================================================================
-DELLY: Structural variant discovery by integrated PE mapping and SR analysis
-============================================================================
-Copyright (C) 2012-2018 Tobias Rausch
-
-This program is free software: you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation, either version 3 of the License, or
-any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <http://www.gnu.org/licenses/>.
-============================================================================
-Contact: Tobias Rausch (rausch@embl.de)
-============================================================================
-*/
-
 #ifndef MODVCF_H
 #define MODVCF_H
 
+#include <boost/progress.hpp>
+
 #include <htslib/sam.h>
 #include <htslib/vcf.h>
 
@@ -44,7 +23,7 @@ void _remove_format_tag(bcf_hdr_t* hdr, bcf1_t* rec, std::string const& tag) {
 }
 
 void _remove_info(bcf_hdr_t* hdr, bcf1_t* rec) {
-  std::string tmp[] = {"CT", "PRECISE", "IMPRECISE", "SVTYPE", "SVMETHOD", "CIEND", "CIPOS", "CHR2", "END", "PE", "MAPQ", "SR", "SRQ", "CONSENSUS"};
+  std::string tmp[] = {"CT", "PRECISE", "IMPRECISE", "SVTYPE", "SVMETHOD", "CIEND", "CIPOS", "CHR2", "POS2", "END", "PE", "MAPQ", "SRMAPQ", "SR", "SRQ", "CONSENSUS"};
   std::set<std::string> keepInfo(tmp, tmp + sizeof(tmp)/sizeof(tmp[0]));
 
   if (!(rec->unpacked & BCF_UN_INFO)) bcf_unpack(rec, BCF_UN_INFO);
@@ -186,6 +165,10 @@ vcfParse(TConfig const& c, bam_hdr_t* hd, std::vector<TStructuralVariantRecord>&
   // Parse bcf
   int32_t nsvend = 0;
   int32_t* svend = NULL;
+  int32_t npos2 = 0;
+  int32_t* pos2 = NULL;
+  int32_t nsvlen = 0;
+  int32_t* svlen = NULL;
   int32_t npe = 0;
   int32_t* pe = NULL;
   int32_t ninslen = 0;
@@ -232,10 +215,15 @@ vcfParse(TConfig const& c, bam_hdr_t* hd, std::vector<TStructuralVariantRecord>&
       svRec.chr = tid;
       svRec.svStart = rec->pos + 1;
       svRec.id = svs.size();
+      svRec.mapq = rec->qual;
       std::string refAllele = rec->d.allele[0];
       std::string altAllele = rec->d.allele[1];
       svRec.alleles = refAllele + "," + altAllele;
 
+      // Parse SV type
+      if ((bcf_get_info_string(hdr, rec, "SVTYPE", &svt, &nsvt) > 0) && (bcf_get_info_string(hdr, rec, "CT", &ct, &nct) > 0)) svRec.svt = _decodeOrientation(std::string(ct), std::string(svt));
+      else continue;
+      
       // Parse INFO
       if (bcf_get_info_flag(hdr, rec, "PRECISE", 0, 0) > 0) svRec.precise=true;
       else svRec.precise = false;
@@ -244,28 +232,56 @@ vcfParse(TConfig const& c, bam_hdr_t* hd, std::vector<TStructuralVariantRecord>&
 	if (svRec.precise) svRec.peSupport = 0;
 	else svRec.peSupport = 2;
       }
-      if (bcf_get_info_int32(hdr, rec, "INSLEN", &inslen, &ninslen) > 0) svRec.insLen = *inslen;
-      else svRec.insLen = 0;
+      if (svRec.svt != 4) {
+	if (bcf_get_info_int32(hdr, rec, "INSLEN", &inslen, &ninslen) > 0) svRec.insLen = *inslen;
+	else svRec.insLen = 0;
+      } else {
+	// Insertions must have INFO/SVLEN
+	if (bcf_get_info_int32(hdr, rec, "SVLEN", &svlen, &nsvlen) > 0) svRec.insLen = *svlen;
+	else continue;
+      }
       if (bcf_get_info_int32(hdr, rec, "HOMLEN", &homlen, &nhomlen) > 0) svRec.homLen = *homlen;
       else svRec.homLen = 0;
       if (bcf_get_info_int32(hdr, rec, "SR", &sr, &nsr) > 0) svRec.srSupport = *sr;
       else svRec.srSupport = 0;
-      if (bcf_get_info_int32(hdr, rec, "END", &svend, &nsvend) > 0) svRec.svEnd = *svend;
-      else svRec.svEnd = rec->pos + 2;
+
+      // SV end assignment
+      svRec.chr2 = tid;
+      svRec.svEnd = rec->pos + 2;
+      if (svRec.svt < DELLY_SVT_TRANS) {
+	// Intra-chromosomal SV
+	if (bcf_get_info_int32(hdr, rec, "END", &svend, &nsvend) > 0) svRec.svEnd = *svend;
+      } else {
+	// Inter-chromosomal SV
+	if (bcf_get_info_string(hdr, rec, "CHR2", &chr2, &nchr2) > 0) {
+	  std::string chr2Name = std::string(chr2);
+	  svRec.chr2 = bam_name2id(hd, chr2Name.c_str());
+	}
+	if (bcf_get_info_int32(hdr, rec, "POS2", &pos2, &npos2) > 0) svRec.svEnd = *pos2;
+      }
       if (bcf_get_info_string(hdr, rec, "CONSENSUS", &cons, &ncons) > 0) svRec.consensus = std::string(cons);
       else svRec.precise = false;
-      if (bcf_get_info_int32(hdr, rec, "CIPOS", &cipos, &ncipos) > 0) svRec.wiggle = cipos[1];
-      else svRec.wiggle = 0;
+      if (bcf_get_info_int32(hdr, rec, "CIPOS", &cipos, &ncipos) > 0) {
+	svRec.ciposlow = cipos[0];
+	svRec.ciposhigh = cipos[1];
+      } else {
+	svRec.ciposlow = -50;
+	svRec.ciposhigh = 50;
+      }
+      if (bcf_get_info_int32(hdr, rec, "CIEND", &cipos, &ncipos) > 0) {
+	svRec.ciendlow = cipos[0];
+	svRec.ciendhigh = cipos[1];
+      } else {
+	svRec.ciendlow = -50;
+	svRec.ciendhigh = 50;
+      }
       if (bcf_get_info_int32(hdr, rec, "MAPQ", &mapq, &nmapq) > 0) svRec.peMapQuality = (uint8_t) *mapq;
       else svRec.peMapQuality = 0;
+      if (bcf_get_info_int32(hdr, rec, "SRMAPQ", &mapq, &nmapq) > 0) svRec.srMapQuality = (uint8_t) *mapq;
+      else svRec.srMapQuality = 0;
       if (bcf_get_info_float(hdr, rec, "SRQ", &srq, &nsrq) > 0) svRec.srAlignQuality = (double) *srq;
       else svRec.srAlignQuality = 0;
-      if ((bcf_get_info_string(hdr, rec, "SVTYPE", &svt, &nsvt) > 0) && (bcf_get_info_string(hdr, rec, "CT", &ct, &nct) > 0)) svRec.svt = _decodeOrientation(std::string(ct), std::string(svt));
-      else continue;
-      if (bcf_get_info_string(hdr, rec, "CHR2", &chr2, &nchr2) > 0) {
-	std::string chr2Name = std::string(chr2);
-	svRec.chr2 = bam_name2id(hd, chr2Name.c_str());
-      } else svRec.chr2 = tid;
+
       svs.push_back(svRec);
     } else if (wimethod == 2) {
       // Assume precise SV, only deletions supported and INFO:END is required!!!
@@ -318,7 +334,10 @@ vcfParse(TConfig const& c, bam_hdr_t* hd, std::vector<TStructuralVariantRecord>&
 	svRec.srSupport = 5;
 	svRec.peMapQuality = 20;
 	svRec.srAlignQuality = 1;
-	svRec.wiggle = 150;
+	svRec.ciposlow = -50;
+	svRec.ciposhigh = 50;
+	svRec.ciendlow = -50;
+	svRec.ciendhigh = 50;
 
 	// Lazy loading of reference sequence
 	if ((seq == NULL) || (tid != lastRefIndex)) {
@@ -353,6 +372,8 @@ vcfParse(TConfig const& c, bam_hdr_t* hd, std::vector<TStructuralVariantRecord>&
   }
   // Clean-up
   free(svend);
+  free(pos2);
+  free(svlen);
   free(svt);
   free(method);
   free(pe);
@@ -404,18 +425,21 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
   bcf_hdr_append(hdr, "##ALT=<ID=INV,Description=\"Inversion\">");
   bcf_hdr_append(hdr, "##ALT=<ID=BND,Description=\"Translocation\">");
   bcf_hdr_append(hdr, "##ALT=<ID=INS,Description=\"Insertion\">");
-  bcf_hdr_append(hdr, "##FILTER=<ID=LowQual,Description=\"PE/SR support below 3 or mapping quality below 20.\">");
+  bcf_hdr_append(hdr, "##FILTER=<ID=LowQual,Description=\"Poor quality and insufficient number of PEs and SRs.\">");
   bcf_hdr_append(hdr, "##INFO=<ID=CIEND,Number=2,Type=Integer,Description=\"PE confidence interval around END\">");
   bcf_hdr_append(hdr, "##INFO=<ID=CIPOS,Number=2,Type=Integer,Description=\"PE confidence interval around POS\">");
-  bcf_hdr_append(hdr, "##INFO=<ID=CHR2,Number=1,Type=String,Description=\"Chromosome for END coordinate in case of a translocation\">");
+  bcf_hdr_append(hdr, "##INFO=<ID=CHR2,Number=1,Type=String,Description=\"Chromosome for POS2 coordinate in case of an inter-chromosomal translocation\">");
+  bcf_hdr_append(hdr, "##INFO=<ID=POS2,Number=1,Type=Integer,Description=\"Genomic position for CHR2 in case of an inter-chromosomal translocation\">");
   bcf_hdr_append(hdr, "##INFO=<ID=END,Number=1,Type=Integer,Description=\"End position of the structural variant\">");
   bcf_hdr_append(hdr, "##INFO=<ID=PE,Number=1,Type=Integer,Description=\"Paired-end support of the structural variant\">");
   bcf_hdr_append(hdr, "##INFO=<ID=MAPQ,Number=1,Type=Integer,Description=\"Median mapping quality of paired-ends\">");
+  bcf_hdr_append(hdr, "##INFO=<ID=SRMAPQ,Number=1,Type=Integer,Description=\"Median mapping quality of split-reads\">");
   bcf_hdr_append(hdr, "##INFO=<ID=SR,Number=1,Type=Integer,Description=\"Split-read support\">");
   bcf_hdr_append(hdr, "##INFO=<ID=SRQ,Number=1,Type=Float,Description=\"Split-read consensus alignment quality\">");
   bcf_hdr_append(hdr, "##INFO=<ID=CONSENSUS,Number=1,Type=String,Description=\"Split-read consensus sequence\">");
   bcf_hdr_append(hdr, "##INFO=<ID=CE,Number=1,Type=Float,Description=\"Consensus sequence entropy\">");
   bcf_hdr_append(hdr, "##INFO=<ID=CT,Number=1,Type=String,Description=\"Paired-end signature induced connection type\">");
+  bcf_hdr_append(hdr, "##INFO=<ID=SVLEN,Number=1,Type=Integer,Description=\"Insertion length for SVTYPE=INS.\">");
   bcf_hdr_append(hdr, "##INFO=<ID=IMPRECISE,Number=0,Type=Flag,Description=\"Imprecise structural variation\">");
   bcf_hdr_append(hdr, "##INFO=<ID=PRECISE,Number=0,Type=Flag,Description=\"Precise structural variation\">");
   bcf_hdr_append(hdr, "##INFO=<ID=SVTYPE,Number=1,Type=String,Description=\"Type of structural variant\">");
@@ -426,10 +450,10 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
   bcf_hdr_append(hdr, "##FORMAT=<ID=GL,Number=G,Type=Float,Description=\"Log10-scaled genotype likelihoods for RR,RA,AA genotypes\">");
   bcf_hdr_append(hdr, "##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">");
   bcf_hdr_append(hdr, "##FORMAT=<ID=FT,Number=1,Type=String,Description=\"Per-sample genotype filter\">");
-  bcf_hdr_append(hdr, "##FORMAT=<ID=RC,Number=1,Type=Integer,Description=\"Raw high-quality read counts for the SV\">");
-  bcf_hdr_append(hdr, "##FORMAT=<ID=RCL,Number=1,Type=Integer,Description=\"Raw high-quality read counts for the left control region\">");
-  bcf_hdr_append(hdr, "##FORMAT=<ID=RCR,Number=1,Type=Integer,Description=\"Raw high-quality read counts for the right control region\">");
-  bcf_hdr_append(hdr, "##FORMAT=<ID=CN,Number=1,Type=Integer,Description=\"Read-depth based copy-number estimate for autosomal sites\">");
+  bcf_hdr_append(hdr, "##FORMAT=<ID=RC,Number=1,Type=Integer,Description=\"Raw high-quality read counts or base counts for the SV\">");
+  bcf_hdr_append(hdr, "##FORMAT=<ID=RCL,Number=1,Type=Integer,Description=\"Raw high-quality read counts or base counts for the left control region\">");
+  bcf_hdr_append(hdr, "##FORMAT=<ID=RCR,Number=1,Type=Integer,Description=\"Raw high-quality read counts or base counts for the right control region\">");
+  bcf_hdr_append(hdr, "##FORMAT=<ID=RDCN,Number=1,Type=Integer,Description=\"Read-depth based copy-number estimate for autosomal sites\">");
   bcf_hdr_append(hdr, "##FORMAT=<ID=DR,Number=1,Type=Integer,Description=\"# high-quality reference pairs\">");
   bcf_hdr_append(hdr, "##FORMAT=<ID=DV,Number=1,Type=Integer,Description=\"# high-quality variant pairs\">");
   if (c.isHaplotagged) {
@@ -458,7 +482,7 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
   // Add samples
   for(unsigned int file_c = 0; file_c < c.files.size(); ++file_c) bcf_hdr_add_sample(hdr, c.sampleName[file_c].c_str());
   bcf_hdr_add_sample(hdr, NULL);
-  bcf_hdr_write(fp, hdr);
+  if (bcf_hdr_write(fp, hdr) != 0) std::cerr << "Error: Failed to write BCF header!" << std::endl;
 
   if (!svs.empty()) {
     // Genotype arrays
@@ -492,15 +516,24 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
     bcf1_t *rec = bcf_init();
     for(typename TSVs::const_iterator svIter = svs.begin(); svIter!=svs.end(); ++svIter) {
       ++show_progress;
+      if ((svIter->srSupport == 0) && (svIter->peSupport == 0)) continue;
+      
       // Output main vcf fields
       int32_t tmpi = bcf_hdr_id2int(hdr, BCF_DT_ID, "PASS");
-      if ((svIter->precise) && (svIter->chr == svIter->chr2)) {
-	if ((svIter->srSupport < 3) || (svIter->peMapQuality < 20)) tmpi = bcf_hdr_id2int(hdr, BCF_DT_ID, "LowQual");
+      if (svIter->chr == svIter->chr2) {
+	// Intra-chromosomal
+	if (((svIter->peSupport < 3) || (svIter->peMapQuality < 20)) && ((svIter->srSupport < 3) || (svIter->srMapQuality < 20))) tmpi = bcf_hdr_id2int(hdr, BCF_DT_ID, "LowQual");
       } else {
-	if ((svIter->peSupport < 3) || (svIter->peMapQuality < 20) || ( (svIter->chr != svIter->chr2) && (svIter->peSupport < 5) ) ) tmpi = bcf_hdr_id2int(hdr, BCF_DT_ID, "LowQual");
+	// Inter-chromosomal
+	if (((svIter->peSupport < 5) || (svIter->peMapQuality < 20)) && ((svIter->srSupport < 5) || (svIter->srMapQuality < 20))) tmpi = bcf_hdr_id2int(hdr, BCF_DT_ID, "LowQual");
       }
       rec->rid = bcf_hdr_name2id(hdr, bamhd->target_name[svIter->chr]);
-      rec->pos = svIter->svStart - 1;
+      int32_t svStartPos = svIter->svStart - 1;
+      if (svStartPos < 1) svStartPos = 1;
+      int32_t svEndPos = svIter->svEnd;
+      if (svEndPos < 1) svEndPos = 1;
+      if (svEndPos >= (int32_t) bamhd->target_len[svIter->chr2]) svEndPos = bamhd->target_len[svIter->chr2] - 1;
+      rec->pos = svStartPos;
       std::string id(_addID(svIter->svt));
       std::string padNumber = boost::lexical_cast<std::string>(svIter->id);
       padNumber.insert(padNumber.begin(), 8 - padNumber.length(), '0');
@@ -517,24 +550,37 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
       std::string dellyVersion("EMBL.DELLYv");
       dellyVersion += dellyVersionNumber;
       bcf_update_info_string(hdr,rec, "SVMETHOD", dellyVersion.c_str());
-      bcf_update_info_string(hdr,rec, "CHR2", bamhd->target_name[svIter->chr2]);
-      tmpi = svIter->svEnd;
-      bcf_update_info_int32(hdr, rec, "END", &tmpi, 1);
+      if (svIter->svt < DELLY_SVT_TRANS) {
+	tmpi = svEndPos;
+	bcf_update_info_int32(hdr, rec, "END", &tmpi, 1);
+      } else {
+	tmpi = svStartPos + 2;
+	bcf_update_info_int32(hdr, rec, "END", &tmpi, 1);
+	bcf_update_info_string(hdr,rec, "CHR2", bamhd->target_name[svIter->chr2]);
+	tmpi = svEndPos;
+	bcf_update_info_int32(hdr, rec, "POS2", &tmpi, 1);
+      }
+      if (svIter->svt == 4) {
+	tmpi = svIter->insLen;
+	bcf_update_info_int32(hdr, rec, "SVLEN", &tmpi, 1);
+      }
       tmpi = svIter->peSupport;
       bcf_update_info_int32(hdr, rec, "PE", &tmpi, 1);
       tmpi = svIter->peMapQuality;
       bcf_update_info_int32(hdr, rec, "MAPQ", &tmpi, 1);
       bcf_update_info_string(hdr, rec, "CT", _addOrientation(svIter->svt).c_str());
       int32_t ciend[2];
-      ciend[0] = -svIter->wiggle;
-      ciend[1] = svIter->wiggle;
+      ciend[0] = svIter->ciendlow;
+      ciend[1] = svIter->ciendhigh;
       int32_t cipos[2];
-      cipos[0] = -svIter->wiggle;
-      cipos[1] = svIter->wiggle;
+      cipos[0] = svIter->ciposlow;
+      cipos[1] = svIter->ciposhigh;
       bcf_update_info_int32(hdr, rec, "CIPOS", cipos, 2);
       bcf_update_info_int32(hdr, rec, "CIEND", ciend, 2);
       
       if (svIter->precise)  {
+	tmpi = svIter->srMapQuality;
+	bcf_update_info_int32(hdr, rec, "SRMAPQ", &tmpi, 1);
 	tmpi = svIter->insLen;
 	bcf_update_info_int32(hdr, rec, "INSLEN", &tmpi, 1);
 	tmpi = svIter->homLen;
@@ -543,9 +589,11 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
 	bcf_update_info_int32(hdr, rec, "SR", &tmpi, 1);
 	float tmpf = svIter->srAlignQuality;
 	bcf_update_info_float(hdr, rec, "SRQ", &tmpf, 1);
-	bcf_update_info_string(hdr, rec, "CONSENSUS", svIter->consensus.c_str());
-	tmpf = entropy(svIter->consensus);
-	bcf_update_info_float(hdr, rec, "CE", &tmpf, 1);
+	if (svIter->consensus.size()) {
+	  bcf_update_info_string(hdr, rec, "CONSENSUS", svIter->consensus.c_str());
+	  tmpf = entropy(svIter->consensus);
+	  bcf_update_info_float(hdr, rec, "CE", &tmpf, 1);
+	}
       }
       
       // Add genotype columns
@@ -603,9 +651,10 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
 	if (gqval[file_c] < 15) ftarr[file_c] = "LowQual";
 	else ftarr[file_c] = "PASS";
       }
-      // ToDo
-      //rec->qual = 0;
-      
+      int32_t qvalout = svIter->mapq;
+      if (qvalout < 0) qvalout = 0;
+      if (qvalout > 10000) qvalout = 10000;
+      rec->qual = qvalout;
       
       bcf_update_genotypes(hdr, rec, gts, bcf_hdr_nsamples(hdr) * 2);
       bcf_update_format_float(hdr, rec, "GL",  gls, bcf_hdr_nsamples(hdr) * 3);
@@ -616,7 +665,7 @@ vcfOutput(TConfig const& c, std::vector<TStructuralVariantRecord> const& svs, TJ
       bcf_update_format_int32(hdr, rec, "RCL", rcl, bcf_hdr_nsamples(hdr));
       bcf_update_format_int32(hdr, rec, "RC", rc, bcf_hdr_nsamples(hdr));
       bcf_update_format_int32(hdr, rec, "RCR", rcr, bcf_hdr_nsamples(hdr));
-      bcf_update_format_int32(hdr, rec, "CN", cnest, bcf_hdr_nsamples(hdr));
+      bcf_update_format_int32(hdr, rec, "RDCN", cnest, bcf_hdr_nsamples(hdr));
       bcf_update_format_int32(hdr, rec, "DR", drcount, bcf_hdr_nsamples(hdr));
       bcf_update_format_int32(hdr, rec, "DV", dvcount, bcf_hdr_nsamples(hdr));
       if (c.isHaplotagged) {
