@@ -1,26 +1,3 @@
-/*
-============================================================================
-DELLY: Structural variant discovery by integrated PE mapping and SR analysis
-============================================================================
-Copyright (C) 2012-2018 Tobias Rausch
-
-This program is free software: you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation, either version 3 of the License, or
-any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <http://www.gnu.org/licenses/>.
-============================================================================
-Contact: Tobias Rausch (rausch@embl.de)
-============================================================================
-*/
-
 #ifndef SPLIT_H
 #define SPLIT_H
 
@@ -145,9 +122,9 @@ namespace torali
   }
 
   // Deletions
-  template<typename TConfig, typename TSeq, typename TSVRecord, typename TRef>
+  template<typename TSeq, typename TSVRecord, typename TRef>
   inline std::string
-  _getSVRef(TConfig const& c, TSeq const* const ref, TSVRecord const& svRec, TRef const refIndex, int32_t const svt) {
+  _getSVRef(TSeq const* const ref, TSVRecord const& svRec, TRef const refIndex, int32_t const svt) {
     if (_translocation(svt)) {
       uint8_t ct = _getSpanOrientation(svt);
       if (svRec.chr==refIndex) {
@@ -191,11 +168,8 @@ namespace torali
       }
     } else {
       if (svt == 2) {
-	if ((c.indels) && (svRec.peSupport == 0) && (svRec.svEnd - svRec.svStart <= c.indelsize)) {
-	  return boost::to_upper_copy(std::string(ref + svRec.svStartBeg, ref + svRec.svEndEnd));
-	} else {
-	  return boost::to_upper_copy(std::string(ref + svRec.svStartBeg, ref + svRec.svStartEnd)) + boost::to_upper_copy(std::string(ref + svRec.svEndBeg, ref + svRec.svEndEnd));
-	}
+	if (svRec.svEnd - svRec.svStart <= DELLY_CHOP_REFSIZE) return boost::to_upper_copy(std::string(ref + svRec.svStartBeg, ref + svRec.svEndEnd));
+	else return boost::to_upper_copy(std::string(ref + svRec.svStartBeg, ref + svRec.svStartEnd)) + boost::to_upper_copy(std::string(ref + svRec.svEndBeg, ref + svRec.svEndEnd));
       } else if (svt == 4) {
 	return boost::to_upper_copy(std::string(ref + svRec.svStartBeg, ref + svRec.svEndEnd));
       } else if (svt == 3) {
@@ -238,9 +212,9 @@ namespace torali
   }
 
 
-  template<typename TConfig, typename TString, typename TSvRecord, typename TAlignDescriptor, typename TPosition>
+  template<typename TString, typename TSvRecord, typename TAlignDescriptor, typename TPosition>
   inline bool
-  _coordTransform(TConfig const& c, TString const& ref, TSvRecord const& sv, TAlignDescriptor const& ad, TPosition& finalGapStart, TPosition& finalGapEnd, int32_t svt) {
+  _coordTransform(TString const& ref, TSvRecord const& sv, TAlignDescriptor const& ad, TPosition& finalGapStart, TPosition& finalGapEnd, int32_t svt) {
     if (_translocation(svt)) {
       uint8_t ct = _getSpanOrientation(svt);
       if (ct == 0) {
@@ -267,14 +241,14 @@ namespace torali
       return true;
     } else {
       if (svt == 2) {
-	if ((c.indels) && (sv.peSupport == 0) && (sv.svEnd - sv.svStart <= c.indelsize)) {
-	  finalGapStart = sv.svStartBeg + ad.rStart;
-	  finalGapEnd = sv.svStartBeg + ad.rEnd;
-	} else {
+	if (sv.svEnd - sv.svStart > DELLY_CHOP_REFSIZE) {
 	  int32_t annealed = sv.svStartEnd - sv.svStartBeg;
 	  if ((ad.rStart >= annealed) || (ad.rEnd < annealed)) return false;
 	  finalGapStart = sv.svStartBeg + ad.rStart;
 	  finalGapEnd = sv.svEndBeg + (ad.rEnd - annealed);
+	} else {
+	  finalGapStart = sv.svStartBeg + ad.rStart;
+	  finalGapEnd = sv.svStartBeg + ad.rEnd;
 	}
 	return true;
       } else if (svt == 3) {
@@ -421,12 +395,11 @@ namespace torali
 
     // Check percent identity
     _percentIdentity(align, gS, gE, ad.percId);
-
     if (ad.percId < c.flankQuality) return false;
 
     // Find homology
     _findHomology(consensus, svRefStr, ad, svt);
-
+    
     // Check flanking alignment length
     if ((ad.homLeft + c.minimumFlankSize > ad.cStart) || ( varIndex < ad.cEnd + ad.homRight + c.minimumFlankSize)) return false;
     if ((ad.homLeft + c.minimumFlankSize > ad.rStart) || ( refIndex < ad.rEnd + ad.homRight + c.minimumFlankSize)) return false;
@@ -438,35 +411,35 @@ namespace torali
   template<typename TAlign>
   inline bool
   _consRefAlignment(std::string const& cons, std::string const& svRefStr, TAlign& aln, int32_t const svt) {
+    AlignConfig<true, false> semiglobal;
+    DnaScore<int> lnsc(5, -4, -4, -4);
+    bool reNeedle = false;
     if (svt == 4) {
-      typedef typename TAlign::index TAIndex;
-      AlignConfig<false, true> semiglobal;
-      DnaScore<int> lnsc(5, -4, -4, -4);
-      bool reNeedle = longNeedle(svRefStr, cons, aln, semiglobal, lnsc);
-      for(TAIndex j = 0; j < (TAIndex) aln.shape()[1]; ++j) {
+      reNeedle = longNeedle(svRefStr, cons, aln, semiglobal, lnsc);
+      for(uint32_t j = 0; j < aln.shape()[1]; ++j) {
 	char tmp = aln[0][j];
 	aln[0][j] = aln[1][j];
 	aln[1][j] = tmp;
       }	
-      return reNeedle;
     } else {
-      AlignConfig<true, false> semiglobal;
-      DnaScore<int> lnsc(5, -4, -4, -4);
-      bool reNeedle = longNeedle(cons, svRefStr, aln, semiglobal, lnsc);
-      return reNeedle;
+      reNeedle = longNeedle(cons, svRefStr, aln, semiglobal, lnsc);
     }
+    return reNeedle;
   }
 
   template<typename TConfig>
   inline bool
   alignConsensus(TConfig const& c, bam_hdr_t* hdr, char const* seq, char const* sndSeq, StructuralVariantRecord& sv) {
-    if ( (int32_t) sv.consensus.size() < (2 * c.minimumFlankSize)) return false;
+    if ( (int32_t) sv.consensus.size() < (2 * c.minimumFlankSize + sv.insLen)) return false;
     
     // Get reference slice
     Breakpoint bp(sv);
-    _initBreakpoint(hdr, bp, sv.consensus.size(), sv.svt);
-    if (bp.chr != bp.chr2) bp.part1 = _getSVRef(c, sndSeq, bp, bp.chr2, sv.svt);
-    std::string svRefStr = _getSVRef(c, seq, bp, bp.chr, sv.svt);
+    if (sv.svt ==4) {
+      int32_t bufferSpace = std::max((int32_t) ((sv.consensus.size() - sv.insLen) / 3), c.minimumFlankSize);
+      _initBreakpoint(hdr, bp, bufferSpace, sv.svt);
+    } else _initBreakpoint(hdr, bp, sv.consensus.size(), sv.svt);
+    if (bp.chr != bp.chr2) bp.part1 = _getSVRef(sndSeq, bp, bp.chr2, sv.svt);
+    std::string svRefStr = _getSVRef(seq, bp, bp.chr, sv.svt);
 
     // Consensus to reference alignment
     typedef boost::multi_array<char, 2> TAlign;
@@ -474,40 +447,60 @@ namespace torali
     if (!_consRefAlignment(sv.consensus, svRefStr, align, sv.svt)) return false;
 
     // Debug consensus to reference alignment
-    //typedef typename TAlign::index TAIndex;
-    //for(TAIndex i = 0; i < (TAIndex) align.shape()[0]; ++i) {
-    //for(TAIndex j = 0; j< (TAIndex) align.shape()[1]; ++j) {
+    //std::cerr << "Consensus-to-Reference alignment" << std::endl;
+    //for(uint32_t i = 0; i < align.shape()[0]; ++i) {
+    //for(uint32_t j = 0; j< align.shape()[1]; ++j) {
     //std::cerr << align[i][j];
     //}
     //std::cerr << std::endl;
     //}
-    //std::cerr << std::endl;
-
+    
     // Check breakpoint
     AlignDescriptor ad;
     if (!_findSplit(c, sv.consensus, svRefStr, align, ad, sv.svt)) return false;
-
+    
     // Get the start and end of the structural variant
     unsigned int finalGapStart = 0;
     unsigned int finalGapEnd = 0;
-    if (!_coordTransform(c, svRefStr, bp, ad, finalGapStart, finalGapEnd, sv.svt)) return false;
-	
+    if (!_coordTransform(svRefStr, bp, ad, finalGapStart, finalGapEnd, sv.svt)) return false;
+    
     sv.precise=true;
     sv.svStart=finalGapStart;
     sv.svEnd=finalGapEnd;
     sv.srAlignQuality = ad.percId;
     sv.insLen=ad.cEnd - ad.cStart - 1;
     sv.homLen=std::max(0, ad.homLeft + ad.homRight - 2);
-    sv.wiggle=std::max(ad.homLeft, ad.homRight);
-    // Set breakpoint, quality and REF & ALT alleles
-    if ((c.indels) && (sv.peSupport == 0) && (sv.svEnd - sv.svStart <= c.indelsize)) {
-      std::string precChar = svRefStr.substr(ad.rStart - 1, 1);
-      std::string refPart = precChar;
-      if (ad.rEnd > ad.rStart + 1) refPart += svRefStr.substr(ad.rStart, (ad.rEnd - ad.rStart) - 1);
-      std::string altPart = precChar;
-      if (ad.cEnd > ad.cStart + 1) altPart += sv.consensus.substr(ad.cStart, (ad.cEnd - ad.cStart) - 1);
-      sv.alleles = refPart + "," + altPart;
+    int32_t ci_wiggle = std::max(ad.homLeft, ad.homRight);
+    sv.ciposlow = -ci_wiggle;
+    sv.ciposhigh = ci_wiggle;
+    sv.ciendlow = -ci_wiggle;
+    sv.ciendhigh = ci_wiggle;
+
+    if (c.islr) {
+      // Set alleles
+      sv.alleles = _addAlleles(boost::to_upper_copy(std::string(seq + sv.svStart - 1, seq + sv.svStart)), std::string(hdr->target_name[sv.chr2]), sv, sv.svt);
+    
+      // Get exact alleles for INS and DEL                                                                         
+      if ((sv.svt == 2) || (sv.svt == 4)) {
+	std::string refVCF;
+	std::string altVCF;
+	int32_t cpos = 0;
+	bool inSV = false;
+	for(uint32_t j = 0; j<align.shape()[1]; ++j) {
+	  if (align[0][j] != '-') {
+	    ++cpos;
+	    if (cpos == ad.cStart) inSV = true;
+	    else if (cpos == ad.cEnd) inSV = false;
+	  }
+	  if (inSV) {
+	    if (align[0][j] != '-') altVCF += align[0][j];
+	    if (align[1][j] != '-') refVCF += align[1][j];
+	  }
+	}
+	sv.alleles = _addAlleles(refVCF, altVCF);
+      }
     }
+
     return true;
   }
 
