@@ -1,26 +1,3 @@
-/*
-============================================================================
-DELLY: Structural variant discovery by integrated PE mapping and SR analysis
-============================================================================
-Copyright (C) 2012-2018 Tobias Rausch
-
-This program is free software: you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation, either version 3 of the License, or
-any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <http://www.gnu.org/licenses/>.
-============================================================================
-Contact: Tobias Rausch (rausch@embl.de)
-============================================================================
-*/
-
 #ifndef TAGS_H
 #define TAGS_H
 
@@ -30,9 +7,17 @@ namespace torali {
   #define DELLY_SVT_TRANS 5
   #endif
 
+  #ifndef DELLY_CHOP_REFSIZE
+  #define DELLY_CHOP_REFSIZE 1000
+  #endif
+
+  #ifndef DELLY_OUTOFBAND
+  #define DELLY_OUTOFBAND -99999999
+  #endif
+
   inline bool
   _translocation(int32_t const svt) {
-    return (DELLY_SVT_TRANS <= svt);
+    return (DELLY_SVT_TRANS <= svt) && (svt < 9);
   }
 
   inline bool
@@ -52,32 +37,41 @@ namespace torali {
 
   // Structural variant record
   struct StructuralVariantRecord {
+    int32_t chr;
     int32_t svStart;
+    int32_t chr2;
     int32_t svEnd;
-    int32_t peSupport;
+    int32_t ciposlow;
+    int32_t ciposhigh;
+    int32_t ciendlow;
+    int32_t ciendhigh;
     int32_t srSupport;
-    int32_t wiggle;
+    int32_t srMapQuality;
+    int32_t mapq;
     int32_t insLen;
-    int32_t homLen;
     int32_t svt;
-    uint32_t id;
+    int32_t id;
+    int32_t homLen;
+    int32_t peSupport;
+    int32_t peMapQuality;
     float srAlignQuality;
     bool precise;
-    uint8_t peMapQuality;
-    int32_t chr;
-    int32_t chr2;
     std::string alleles;
     std::string consensus;
 
-  StructuralVariantRecord() : svStart(0), svEnd(0), peSupport(0), srSupport(0), wiggle(0), insLen(0), homLen(0), svt(-1), id(0), srAlignQuality(0), precise(false), peMapQuality(0), chr(0), chr2(0) {}
-  StructuralVariantRecord(int32_t const c, int const s, int const e) : svStart(s), svEnd(e), peSupport(0), srSupport(0), wiggle(0), insLen(0), homLen(0), svt(-1), id(0), srAlignQuality(0), precise(false), peMapQuality(0), chr(c), chr2(c) {}
+    
+    StructuralVariantRecord() : chr(0), svStart(0), chr2(0), svEnd(0), ciposlow(0), ciposhigh(0), ciendlow(0), ciendhigh(0), srSupport(0), srMapQuality(0), mapq(0), insLen(0), svt(-1), id(0), homLen(0), peSupport(0), peMapQuality(0), srAlignQuality(0), precise(false) {}
+
+    StructuralVariantRecord(int32_t const c, int32_t const s, int32_t const e) : chr(c), svStart(s), chr2(c), svEnd(e), ciposlow(0), ciposhigh(0), ciendlow(0), ciendhigh(0), srSupport(0), srMapQuality(0), mapq(0), insLen(0), svt(-1), id(0), homLen(0), peSupport(0), peMapQuality(0), srAlignQuality(0), precise(false) {}
+
+    StructuralVariantRecord(int32_t const c1, int32_t const s, int32_t const c2, int32_t const e, int32_t const cipl, int32_t const ciph, int32_t const ciel, int32_t const cieh, int32_t const sup, int32_t const srmapq, int32_t const qval, int32_t const ilen, int32_t const svtype, int32_t const idval): chr(c1), svStart(s), chr2(c2), svEnd(e), ciposlow(cipl), ciposhigh(ciph), ciendlow(ciel), ciendhigh(cieh), srSupport(sup), srMapQuality(srmapq), mapq(qval), insLen(ilen), svt(svtype), id(idval), homLen(0), peSupport(0), peMapQuality(0), srAlignQuality(0), precise(true) {}
   };
 
   template<typename TSV>
   struct SortSVs : public std::binary_function<TSV, TSV, bool>
   {
     inline bool operator()(TSV const& sv1, TSV const& sv2) {
-      return ((sv1.chr<sv2.chr) || ((sv1.chr==sv2.chr) && (sv1.svStart<sv2.svStart)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.svEnd<sv2.svEnd)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.svEnd==sv2.svEnd) && (sv1.peSupport > sv2.peSupport)));
+      return ((sv1.chr<sv2.chr) || ((sv1.chr==sv2.chr) && (sv1.svStart<sv2.svStart)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.chr2<sv2.chr2)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.chr2==sv2.chr2) && (sv1.svEnd<sv2.svEnd)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.chr2==sv2.chr2) && (sv1.svEnd==sv2.svEnd) && (sv1.peSupport > sv2.peSupport)) || ((sv1.chr==sv2.chr) && (sv1.svStart==sv2.svStart) && (sv1.chr2==sv2.chr2) && (sv1.svEnd==sv2.svEnd) && (sv1.peSupport == sv2.peSupport) && (sv1.srSupport > sv2.srSupport)));
     }
   };
 
@@ -154,47 +148,62 @@ namespace torali {
   template<typename TSize>
   inline bool
   _svSizeCheck(TSize const s, TSize const e, int32_t const svt) {
-    if (svt == 0) return (( e - s ) >= 100);
-    else if (svt == 1) return (( e - s ) >= 100);
+    // Short reads
+    if (svt == 0) return (( e - s ) >= 300);
+    else if (svt == 1) return (( e - s ) >= 300);
     else if (svt == 2) return (( e - s ) >= 300);
     else if (svt == 3) return (( e - s ) >= 100);
     else return true;
   }
 
+  template<typename TSize>
+  inline bool
+  _svSizeCheck(TSize const s, TSize const e, int32_t const svt, int32_t const inslen) {
+    // Long reads
+    if (svt == 0) return (( e - s ) >= 15);
+    else if (svt == 1) return (( e - s ) >= 15);
+    else if (svt == 2) return (( e - s ) >= 15);
+    else if (svt == 3) return (( e - s ) >= 15);
+    else if (svt == 4) return (inslen >= 15);
+    else return true;
+  }
+
 
   // 0: Left-spanning inversion
   // 1: Right-spanning inversion
   // 2: Deletion-type
   // 3: Duplication-type
 
-  template<typename TBamRecord>
   inline uint8_t
-  getSVType(TBamRecord const& al) {
-    if (al.flag & BAM_FREAD1) {
-      if (!(al.flag & BAM_FREVERSE)) {
-	if (!(al.flag & BAM_FMREVERSE)) return 0;
-	else return (al.pos < al.mpos) ? 2 : 3;
-      } else {
-	if (!(al.flag & BAM_FMREVERSE)) return (al.pos > al.mpos) ? 2 : 3;
-	else return 1;
-      }
+  getSVType(bam1_t* rec) {
+    if (!(rec->core.flag & BAM_FREVERSE)) {
+      if (!(rec->core.flag & BAM_FMREVERSE)) return 0;
+      else return (rec->core.pos < rec->core.mpos) ? 2 : 3;
     } else {
-      if (!(al.flag & BAM_FREVERSE)) {
-	if (!(al.flag & BAM_FMREVERSE)) return 0;
-	else return (al.pos < al.mpos) ? 2 : 3;
-      } else {
-	if (!(al.flag & BAM_FMREVERSE)) return (al.pos > al.mpos) ? 2 : 3;
-	else return 1;
-      }
+      if (!(rec->core.flag & BAM_FMREVERSE)) return (rec->core.pos > rec->core.mpos) ? 2 : 3;
+      else return 1;
     }
   }
 
   inline int32_t
   _isizeMappingPos(bam1_t* rec, int32_t isize) {
-    if (_translocation(rec)) return DELLY_SVT_TRANS + getSVType(rec->core);
-    else {
+    if (_translocation(rec)) {
+      uint8_t orient = getSVType(rec);
+      if (orient == 0) return DELLY_SVT_TRANS + 0;
+      else if (orient == 1) return DELLY_SVT_TRANS + 1;
+      else {
+	// 3to5 or 5to3?
+	if (rec->core.tid > rec->core.mtid) {
+	  if (!(rec->core.flag & BAM_FREVERSE)) return DELLY_SVT_TRANS + 2;
+	  else return DELLY_SVT_TRANS + 3;
+	} else {
+	  if (!(rec->core.flag & BAM_FREVERSE)) return DELLY_SVT_TRANS + 3;
+	  else return DELLY_SVT_TRANS + 2;
+	}
+      }
+    } else {
       if (rec->core.pos == rec->core.mpos) return -1; // No SV
-      uint8_t orient = getSVType(rec->core);
+      uint8_t orient = getSVType(rec);
       if (orient == 0) return 0;
       else if (orient == 1) return 1;
       else if (orient == 2) {
@@ -204,24 +213,6 @@ namespace torali {
 	if (std::abs(rec->core.pos - rec->core.mpos) < 100) return -1; // Too small
 	return 3;
       }
-    } 
-  }
-
-  // Decode Orientation
-  inline int32_t
-  _decodeOrientation(std::string const& value, std::string const& svt) {
-    if (svt == "BND") {
-      if (value=="3to3") return DELLY_SVT_TRANS + 0;
-      else if (value=="5to5") return DELLY_SVT_TRANS + 1;
-      else if (value=="3to5") return DELLY_SVT_TRANS + 2;
-      else if (value=="5to3") return DELLY_SVT_TRANS + 3;
-      else return -1;
-    } else {
-      if (value=="3to3") return 0;
-      else if (value=="5to5") return 1;
-      else if (value=="3to5") return 2;
-      else if (value=="5to3") return 3;
-      else return 4;
     }
   }
 
@@ -241,21 +232,6 @@ namespace torali {
     else return (rec->core.tid < rec->core.mtid);
   }
 
-  template<typename TLibraryMap>
-  inline int32_t
-  _findLib(bam1_t* rec, TLibraryMap& lib) {
-    std::string rG = "DefaultLib";
-    uint8_t *rgptr = bam_aux_get(rec, "RG");
-    if (rgptr) {
-      char* rg = (char*) (rgptr + 1);
-      rG = std::string(rg);
-    }
-    for(uint32_t libIdx = 0; libIdx < lib.size(); ++libIdx)
-      if (lib[libIdx].rg == rG) return libIdx;
-    return 0;
-  }
-  
-
   // Deletions
   template<typename TSize, typename TISize>
   inline bool
